/************************************************************************/
/*																		*/
/*	main.c	--	RTOS with USB Host to write to ThumbDrive
 *                      Also demonstrates use of RTOS queue and how to check task stack usage and heap usage
 *
 *      IMPORTANT: MX4cK JP6 must be in HOST position, J16 on VBUSON, JP10 Jumpered
 *
 *      Program sets up two threads.
 *      doTask flashed an LED and waits for user to press the built in button 1.  It then allocates space on the heap for
 *      the data and fills in dummy data.  Finally it pushes a pointer to the data on to a queue.
 *      The queue is only one item deep, so the queue blocks if there is not room
 *
 *      usbTask monitors the USB device.  Once the device is available it checks the queue for data
 *      If data is  available it pops the data off of the queue and writes the data to the usb drive
 *
 *      The USB handling uses an interrupt. Since it does not use any RTOS features
 *      it is not wrapped in an assembly wrapper for the RTOS.  See usb_host.c for notes
 *      on wrapping the ISR 
 *
 *      Program turns on PORT_B bit 13 (built in) if RTOS error occurs
 *
 * Based on FreeRTOS V7.3.0 - Copyright (C) 2012 Real Time Engineers Ltd.
 *
 * Mason, Seattle U
 *      Revised 05/08/2013
/*																		*/
/************************************************************************/
#include "MX4cKsetup.h"
#include "MX4cKsettings.h"

#include "RTOS/FreeRTOS.h"
#include "RTOS/task.h"
#include "RTOS/queue.h"
#include "RTOS/timers.h"
#include "RTOS/semphr.h"
#include "RTOS/queue.h"

#include "USBMSD/usb.h"
#include "USBMSD/FSIO.h"

// forward declaration
static void SetupHardware(void);
static void usbTask(void *pvParameters);
static void doTask(void *pvParameters);

xTaskHandle usbTaskHandle;  // holds a handle to the USB task
                            // only needed because we check task memory usage in this example

// create a queue to pass data to the USB thumb drive thread
xQueueHandle dataQueue;

// structure for holding data written to the drive
typedef struct {
    char filename[20];
    BYTE data[100]; // data is held here
    int n; //data size
} DATA;

// globally holds size of DATA
int sizeofDATA;

// USB global flag
volatile BOOL deviceAttached;

int main(void) {

    // size of the data set structure
    // called in main before the RTOS is started. sizeof was not working correctly from inside a thread
    sizeofDATA = sizeof (DATA);

    // Prepare the hardware to run
    SetupHardware();
    deviceAttached = FALSE; // initialize USB state

    // set the digital io ports
    // configure the built-in LEDs
    //   LED1 is on RB10
    //	 LED2 is on RB11
    //	 LED3 is on RB12
    //	 LED4 is on RB13
    //Set ports for onboard LEDs to outputs
    PORTSetPinsDigitalOut(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13);
    PORTClearBits(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13); // turn off the LEDs

    // set the built-in buttons as inputs
    PORTSetPinsDigitalIn(IOPORT_A, BIT_6 | BIT_7);

    // create a queue,  1 item, 4 bytes (a pointer)
    dataQueue = xQueueCreate(1, 4);

    // task to collect and process data
    xTaskCreate(doTask, // The function that implements the task.
            (signed char *) "DATA", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE+250, // Extra space for USB function
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            1, // The priority assigned to the task. Currently lower than the timer
            NULL); // The task handle is not required, so NULL is passed.



    // thread to handle USB and file system
    xTaskCreate(usbTask, // The function that implements the task.
            (signed char *) "USB", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE+250, // Extra space for USB function 
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            1, // The priority assigned to the task. Currently lower than the timer
            &usbTaskHandle); // The task handle is not required, so NULL is passed.



    /* Start the tasks and timer running. */
    vTaskStartScheduler();

    for (;;);
}

//-----------------------------------------------------------
// thread to handle data processing
// thread monitors internal button #1.  When pressed it pushes dummy data on the queue
// The USB thread monitors the queue, pulls data from the queue and writes it to a file on the USB drive
static void doTask(void *pvParameters) {
    
    int i;
    char fileid = '0';      // uncremented everytime new data is pushed on the queue.  Used to create unique filenames.
    DATA *theData;          // pointer to data to be written to the USB drive.
   
    
    for (;;) {
       
        //  wait until the button is pressed
        // flash the led while waiting
        while (!PORTReadBits(IOPORT_A, BIT_6)) {
            PORTToggleBits(IOPORT_B, BIT_12); // turn off the LEDs
            vTaskDelay(500);  // makes for poor button response but simple code
                              // should check for change in button state and exit loop then
        }

        // use pressed the button
        // allocate space on the heap for the data set
        // fill the data
        // push it to the queue
        // NOTE: the RTOS is configured using heap_2.c  This memory management option
        // does not combine free blocks.  As long as you allocate and free blocks that are the
        // same size the heap will not grow.  If you modify this program to use an arbitrary size
        // data structure you need to statically allocate memory or use one of the other memory management
        // options in freeRTOS
        theData = pvPortMalloc(sizeofDATA);

        // create a unique file name  - at least unique to this program this times its run
        strcpy(theData->filename, "data0.txt");
        theData->filename[4] = fileid++;  // change "0" to next letter.
                                          // this may fail when the ascii chararacter exceeds '9'

        // fill the structure with dummyfreestack = uxTaskGetStackHighWaterMark(usbTaskHandle); data
       
        theData->n =sprintf(theData->data,"USB Task Stack Remaining: %d, RTOS Heap Remaining: %d",
                     uxTaskGetStackHighWaterMark(usbTaskHandle),
                     xPortGetFreeHeapSize())+1;

        // push pointer to data onto the queue
        // the queue length is set to one, so if the USB drive has not pulled the previous
        // data set from the queue this function blocks until the queue is empty
        xQueueSendToBack(dataQueue, &theData, portMAX_DELAY);

        while (PORTReadBits(IOPORT_A, BIT_6)); // wait for button release to avoid multiple writes per button press
        vTaskDelay(500);                        // primitive debounce

    }
}
//-----------------------------------------------------------
// thread to handle the usb drive
// this thread must be called periodically
// the thread waits until the USB drive is available and data is available on the queue
// it then pulls the data and writes it to the USB drive
static void usbTask(void *pvParameters) {

    DATA *theData;          // pointer to data set
    FSFILE * myFile;        // the file

    //Initialize the USB stack
    USBInitialize(0);
  
    for (;;) {
  
        USBTasks(); // loop until the drive is plugged in

        //if thumbdrive is plugged in
        if (USBHostMSDSCSIMediaDetect()) {
            PORTSetBits(IOPORT_B, BIT_10); // USB device recognized    
            deviceAttached = TRUE;
        } else if (!deviceAttached) {
            PORTClearBits(IOPORT_B, BIT_10); // USB device recognized   
        }



        if ((deviceAttached) && xQueueReceive(dataQueue, &theData, 0)) {

            // if the thumb drive is plugged in and there is data in the queue
            // then pull the pointer the data off of the queue and write the data to the USB drive

            //Initialize the filesystem
            PORTSetBits(IOPORT_B, BIT_11);  // turn on this LED while writing data
                                            // this may be a fast flash depending on data size
            if (FSInit()) {
                // the file system is initialize before every write.  This is not efficient.  Better coding
                // to only initialize the file system if the device changes state from Detached to Attached

                // Opening a file in mode "w" will create the file if it doesn't
                //  exist.  If the file does exist it will delete the old file
                //  and create a new one that is blank.
                myFile = FSfopen(theData->filename, "w");

                //Write the data to the new file.
                FSfwrite(theData->data, 1, theData->n, myFile);

                //Always make sure to close the file so that the data gets
                //  written to the drive.
                FSfclose(myFile);

            }
            // all done with data so free memory
            vPortFree(theData);
            PORTClearBits(IOPORT_B, BIT_11); // turn off LED when writing is done

        }


    }

}

static void SetupHardware(void) {
    /* Set the system and peripheral bus speeds and enable the program cache*/
    SYSTEMConfigPerformance(configCPU_CLOCK_HZ); // was  configCPU_CLOCK_HZ-1
    mOSCSetPBDIV(OSC_PB_DIV_8);

    /* Setup to use the external interrupt controller. */
    INTEnableSystemMultiVectoredInt();

    portDISABLE_INTERRUPTS();
}

/*-----------------------------------------------------------
 *  Hooks for RTOS.  Are not required to modify these
 *  May want to turn on LED or some other external signal if an error
 *  occurs
 *-----------------------------------------------------------*/

void vApplicationMallocFailedHook(void) {
    // support function.  See RTOS Starter
    // if an error occurs go into an infinite loop
    taskDISABLE_INTERRUPTS();
    PORTSetBits(IOPORT_B, BIT_13);
    for (;;);
}

void vApplicationIdleHook(void) {
    // support function.  See RTOS Starter
}

void vApplicationStackOverflowHook(xTaskHandle pxTask, signed char *pcTaskName) {
    (void) pcTaskName;
    (void) pxTask;
    // support function.  See RTOS Starter
    // if an error occurs go into an infinite loop
    taskDISABLE_INTERRUPTS();
    for (;;);
}

void vApplicationTickHook(void) {
    // support function.  See RTOS Starter
}

void _general_exception_handler(unsigned long ulCause, unsigned long ulStatus) {
    // support function.  See RTOS Starter
    // if an error occurs go into an infinite loop
    PORTSetBits(IOPORT_B, BIT_13);
    for (;;);
}

/****************************************************************************
  USB Hook so can respond to USB events
  This function does not need to be modified

  Function:
    BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event,
                void *data, DWORD size )

  Summary:
    This is the application event handler.  It is called when the stack has
    an event that needs to be handled by the application layer rather than
    by the client driver.

  Description:
    This is the application event handler.  It is called when the stack has
    an event that needs to be handled by the application layer rather than
    by the client driver.  If the application is able to handle the event, it
    returns TRUE.  Otherwise, it returns FALSE.

  Precondition:
    None

  Parameters:
    BYTE address    - Address of device where event occurred
    USB_EVENT event - Identifies the event that occured
    void *data      - Pointer to event-specific data
    DWORD size      - Size of the event-specific data

  Return Values:
    TRUE    - The event was handled
    FALSE   - The event was not handled

  Remarks:
    The application may also implement an event handling routine if it
    requires knowledge of events.  To do so, it must implement a routine that
    matches this function signature and define the USB_HOST_APP_EVENT_HANDLER
    macro as the name of that function.
 ***************************************************************************/

BOOL USB_ApplicationEventHandler(BYTE address, USB_EVENT event, void *data, DWORD size) {
    switch (event) {
        case EVENT_VBUS_REQUEST_POWER:
            // The data pointer points to a byte that represents the amount of power
            // requested in mA, divided by two.  If the device wants too much power,
            // we reject it.
            return TRUE;

        case EVENT_VBUS_RELEASE_POWER:
            // Turn off Vbus power.
            // The PIC24F with the Explorer 16 cannot turn off Vbus through software.

            //This means that the device was removed
            deviceAttached = FALSE;
            return TRUE;
            break;

        case EVENT_HUB_ATTACH:
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            //UART2PrintString( "\r\n***** USB Error - cannot enumerate device *****\r\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            //UART2PrintString( "\r\n***** USB Error - client driver initialization error *****\r\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            //UART2PrintString( "\r\n***** USB Error - out of heap memory *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR: // This should never be generated.
            //UART2PrintString( "\r\n***** USB Error - unspecified *****\r\n" );
            return TRUE;
            break;

        default:
            break;
    }

    return FALSE;
}
